#include <linux/delay.h>
#include <asm/uaccess.h>
#include "mod_utils.h"
#include "linux/cpld-usr.h"
#include "cpld_internal.h"

/* We don't want to ever accidently timeout.  The WishBone peripheral will
 * wait for the ACK when reading from a slave, so reads can be variable
 * length.  95% of transactions take <1ms, while ~5% have been seen to take
 * ~30ms and ~1% take ~65ms.  The hardware timeout risks locking up the
 * WishBone i2c perhipheral, since it may stop in the middle of a
 * slave-to-master transfer.  The WishBone interface will then see all
 * future transactions as an "arbitration loss", since the slave is still
 * driving SDA low.  It is normally easy to recover from this by simply
 * driving SCL 9 times, but the WishBone peripheral refuses to do so in this
 * case and cannot be recovered without a power cycle. */
#define TIMEOUT_MS 500
/* I2C_C5 timeout is in increments of 56ms. Give at least 50ms headroom. */
#define REG_I2C_C5_TIMEOUT (((TIMEOUT_MS-50)/56-1)<<4)

/* prevent concurrent access to i2c registers and the cache */
static DEFINE_MUTEX(cpld_sfp_lock);

// this structure holds the data that is captured on good i2c reads
// when i2c loses arbitration, which makes i2c read impossible this data is returned instead
static struct {
   DECLARE_BITMAP(valid, 256);
   u8 data[256];
   unsigned used; /* how many times cache was used */
} sfp_id_cache;

static int cpld_sfp_poll_ready(void)
{
   uint8_t val;
   int err;
   unsigned long timeout = jiffies + msecs_to_jiffies(TIMEOUT_MS);

   for (;;) {
      /* i2c_s1[READY] is not set for writes, so we need to check the debug register */
      val = readb_cpld_i2c_dbg0();
      if ((val & CPLD_I2C_DBG0_STATE_MASK) == CPLD_I2C_DBG0_STATE_RDY) {
         val = readb_cpld_i2c_s1();
         if (val & CPLD_I2C_S1_ACCESS_BAD) {
            err = -EIO;
            break;
         }

         return 0;
      }

      if (time_is_before_jiffies(timeout)) {
         pr_err("i2c_s1 software timeout expired\n");
         err = -EBUSY;
         break;
      }

      msleep(1);
   }

   /* error case: reset the HW here to prevent bus lockup */
   writeb_cpld_i2c_c5(REG_I2C_C5_TIMEOUT | CPLD_I2C_C5_RESET);
   writeb_cpld_i2c_1_cr(0);
   writeb_cpld_i2c_1_cr(0x80);
   writeb_cpld_i2c_c5(REG_I2C_C5_TIMEOUT);

   return err;
}

#define SFP_BUS_BH      0x00
#define SFP_BUS_LMTS0   0x10
#define SFP_BUS_LMTS1   0x30

#define SFP_DEV_ID      0x50
#define SFP_DEV_DIAG    0x51

static int cpld_sfp_read_byte(unsigned bus, unsigned dev, u8 addr)
{
   int err;

   writeb_cpld_i2c_c3(addr);
   writeb_cpld_i2c_c1(bus | ((dev==SFP_DEV_ID) ? CPLD_I2C_C1_ID_READ : CPLD_I2C_C1_DIAG_READ));
   err = cpld_sfp_poll_ready();
   if (err)
      return err;
   return readb_cpld_i2c_s2();
}

static int cpld_sfp_read(unsigned bus, unsigned dev, u8 addr, u8 *data, unsigned count)
{
   int err;

   mutex_lock(&cpld_sfp_lock);

   while (count--) {
      err = cpld_sfp_read_byte(bus, dev, addr++);
      if (err < 0)
         goto out;
      *data++ = err;
   }

   err = 0;

out:
   mutex_unlock(&cpld_sfp_lock);
   return err;
}

static int cpld_sfp_write_byte(unsigned bus, unsigned dev, u8 addr, u8 data)
{
   if (dev != SFP_DEV_DIAG)
      return -EROFS;

   writeb_cpld_i2c_c3(addr);
   writeb_cpld_i2c_c4(data);
   writeb_cpld_i2c_c1(bus | CPLD_I2C_C1_DIAG_WRITE);
   return cpld_sfp_poll_ready();
}

static int cpld_sfp_write(unsigned bus, unsigned dev, u8 addr, const u8 *data, unsigned count)
{
   int err;

   mutex_lock(&cpld_sfp_lock);

   while (count--) {
      err = cpld_sfp_write_byte(bus, dev, addr++, *data++);
      if (err)
         goto out;
   }

out:
   mutex_unlock(&cpld_sfp_lock);
   return err;
}

static bool bitmap_is_range_set(const unsigned long *addr, unsigned off, unsigned len)
{
   unsigned end = off + len;
   return end == find_next_zero_bit(sfp_id_cache.valid, end, off);
}

static int cpld_sfp_serial_read(struct cpld_sfp_serial_read __user *usr)
{
   struct cpld_sfp_serial_read sfp_read;
   int ret;
   u8 addr, len;

   if (copy_from_user(&sfp_read, usr, sizeof(sfp_read))) {
      MOD_ERROR("copy_from_user failed\n");
      return -EFAULT;
   }

   switch (sfp_read.attr) {
   case SFP_VENDOR_NAME:
      addr = 20;
      len = 16;
      break;

   case SFP_PART_NUMBER:
      addr = 40;
      len = 16;
      break;

   case SFP_DATE_CODE:
      addr = 84;
      len = 8;
      break;

   case SFP_SERIAL_NUMBER:
      addr = 68;
      len = 16;
      break;

   case SFP_VENDOR_REVISION:
      addr = 56;
      len = 4;
      break;

   case SFP_WAVE_LENGTH:
      addr = 60;
      len = 2;
      break;

   case SFP_SFF_8472_COMPLIANCE:
      addr = 94;
      len = 1;
      break;

   case SFP_TRANCEIVER_CODE:
      addr = 3;
      len = 8;
      break;

   case SFP_CONNECTOR_VALUE:
      addr = 2;
      len = 1;
      break;

   case SFP_DIAG_MONITOR_TYPE:
      addr = 92;
      len = 1;
      break;

   case SFP_BR_NOMINAL:
      addr = 12;
      len = 1;
      break;

   case SFP_BR_MIN:
      addr = 67;
      len = 1;
      break;

   default:
      return -EINVAL;
   }

   ret = cpld_sfp_read(SFP_BUS_BH, SFP_DEV_ID, addr, sfp_read.regs, len);
   mutex_lock(&cpld_sfp_lock);
   if (ret) {
      /* error; try to use the cached copy */
      if (bitmap_is_range_set(sfp_id_cache.valid, addr, len)) {
         memcpy(sfp_read.regs, &sfp_id_cache.data[addr], len);
         sfp_id_cache.used++;
      } else {
         mutex_unlock(&cpld_sfp_lock);
         return ret;
      }
   } else {
      /* save the cache */
      memcpy(&sfp_id_cache.data[addr], sfp_read.regs, len);
      bitmap_set(sfp_id_cache.valid, addr, len);
   }
   mutex_unlock(&cpld_sfp_lock);

   if (copy_to_user(usr, &sfp_read, sizeof(sfp_read))) {
      MOD_ERROR("copy_to_user failed\n");
      return -EFAULT;
   }
   return 0;
}

static int cpld_sfp_diag_read(struct cpld_sfp_diag_read __user *usr)
{
   struct cpld_sfp_diag_read sfp_diag;
   int ret;
   u8 addr, len;

   if (copy_from_user(&sfp_diag, usr, sizeof(sfp_diag))) {
      MOD_ERROR("copy_from_user failed\n");
      return -EFAULT;
   }

   switch (sfp_diag.diagAttr) {
   case SFP_DIAG_STATUS_CTRL:
      addr = 110;
      len = 1;
      break;

   case SFP_DIAG_EXT_STATUS_CTRL:
      addr = 118;
      len = 1;
      break;

   case SFP_DIAG_TEMP:
      addr = 96;
      len = 2;
      break;

   case SFP_DIAG_VOLTAGE:
      addr = 98;
      len = 2;
      break;

   case SFP_DIAG_CURRENT:
      addr = 100;
      len = 2;
      break;

   case SFP_DIAG_TX_POWER:
      addr = 102;
      len = 2;
      break;

   case SFP_DIAG_RX_POWER:
      addr = 104;
      len = 2;
      break;

   case SFP_DIAG_TSLOPE:
      addr = 84;
      len = 2;
      break;

   case SFP_DIAG_TOFFSET:
      addr = 86;
      len = 2;
      break;

   case SFP_DIAG_VSLOPE:
      addr = 88;
      len = 2;
      break;

   case SFP_DIAG_VOFFSET:
      addr = 90;
      len = 2;
      break;

   case SFP_DIAG_ISLOPE:
      addr = 76;
      len = 2;
      break;

   case SFP_DIAG_IOFFSET:
      addr = 78;
      len = 2;
      break;

   case SFP_DIAG_TX_PWR_SLOPE:
      addr = 80;
      len = 2;
      break;

   case SFP_DIAG_TX_PWR_OFFSET:
      addr = 82;
      len = 2;
      break;

   case SFP_DIAG_RX_PWR_0:
      addr = 72;
      len = 4;
      break;

   case SFP_DIAG_RX_PWR_1:
      addr = 68;
      len = 4;
      break;

   case SFP_DIAG_RX_PWR_2:
      addr = 64;
      len = 4;
      break;

   case SFP_DIAG_RX_PWR_3:
      addr = 60;
      len = 4;
      break;

   case SFP_DIAG_RX_PWR_4:
      addr = 56;
      len = 4;
      break;

   case SFP_DIAG_ALARM_WARNING_BYTES:
      addr = 112;
      len = 4;
      break;

   default:
      return -EINVAL;
   }

   ret = cpld_sfp_read(SFP_BUS_BH, SFP_DEV_DIAG, addr, sfp_diag.regs, len);
   if (ret)
      return ret;

   if (copy_to_user(usr, &sfp_diag, sizeof(sfp_diag))) {
      MOD_ERROR("copy_to_user failed\n");
      return -EFAULT;
   }
   return 0;
}


static int cpld_sfp_diag_write_set(u8 addr, u8 value)
{
   return cpld_sfp_write(SFP_BUS_BH, SFP_DEV_DIAG, addr, &value, 1);
}


static int cpld_sfp_diag_write(struct cpld_sfp_diag_write __user *usr)
{
   struct cpld_sfp_diag_write sfp_diag;

   if (copy_from_user(&sfp_diag, usr, sizeof(sfp_diag))) {
      MOD_ERROR("copy_from_user failed\n");
      return -EFAULT;
   }

   switch (sfp_diag.diagAttr) {
   case SFP_DIAG_STATUS_CTRL:
      return cpld_sfp_diag_write_set(110, sfp_diag.regs);

   case SFP_DIAG_EXT_STATUS_CTRL:
      return cpld_sfp_diag_write_set(118, sfp_diag.regs);

   default:
      return -EINVAL;
   }

   return 0;
}

long cpld_sfp_ioctl(struct file *file, unsigned int cmd, unsigned long arg)
{
   void __user *usr_ptr = (void __user *)arg;
   uint8_t val = 0;
   int ret = 0;

   switch (cmd) {
   case CPLD_SFP_SERIAL_READ:
      return cpld_sfp_serial_read(usr_ptr);

   case CPLD_SFP_SERIAL_WRITE:
      return -EROFS; /* this should never work since ID is r/o */

   case CPLD_SFP_DIAG_READ:
      return cpld_sfp_diag_read(usr_ptr);

   case CPLD_SFP_DIAG_WRITE:
      return cpld_sfp_diag_write(usr_ptr);

   case CPLD_SFP_CTRL_READ:
      cpld_lock();
      val = readb_cpld_sfpc();
      cpld_unlock();
      if (copy_to_user(usr_ptr, &val, sizeof(val))) {
         MOD_ERROR("CPLD_SFP_CTRL_READ: copy_to_user failed\n");
         ret = -EFAULT;
      }
      break;

   case CPLD_SFP_CTRL_WRITE: {
      struct cpld_sfpc_write sfpc_write;

      if (copy_from_user(&sfpc_write, usr_ptr, sizeof(sfpc_write))) {
         MOD_ERROR("CPLD_SFP_CTRL_WRITE: copy_from_user failed\n");
         ret = -EFAULT;
      }
      cpld_lock();
      writeb_cpld_sfpc(sfpc_write.regs);
      cpld_unlock();
      break;

   }
   case CPLD_SFP_SWSTAT_READ:
      cpld_lock();
      val = readb_cpld_swstat();
      cpld_unlock();
      if (copy_to_user(usr_ptr, &val, sizeof(val))) {
         MOD_ERROR("CPLD_SFP_SWSTAT_READ: copy_to_user failed\n");
         ret = -EFAULT;
      }
      break;
   default:
      MOD_ERROR("invalid ioctl: %u\n", cmd);
      ret = -ENOTTY;
   }

   return ret;
}

static ssize_t bh_sfp_id_read(struct file *filp, struct kobject *kobj,
                              struct bin_attribute *bin_attr,
                              char *buffer, loff_t offset, size_t count)
{
   /* setting the bin_attribute size handles offset and count limits */
   int ret = cpld_sfp_read(SFP_BUS_BH, SFP_DEV_ID, offset, buffer, count);
   return ret ? : count;
}

static ssize_t bh_sfp_diag_read(struct file *filp, struct kobject *kobj,
                              struct bin_attribute *bin_attr,
                              char *buffer, loff_t offset, size_t count)
{
   /* setting the bin_attribute size handles offset and count limits */
   int ret = cpld_sfp_read(SFP_BUS_BH, SFP_DEV_DIAG, offset, buffer, count);
   return ret ? : count;
}

static ssize_t bh_sfp_diag_write(struct file *filp, struct kobject *kobj,
                              struct bin_attribute *bin_attr,
                              char *buffer, loff_t offset, size_t count)
{
   /* setting the bin_attribute size handles offset and count limits */
   int ret = cpld_sfp_write(SFP_BUS_BH, SFP_DEV_DIAG, offset, buffer, count);
   return ret ? : count;
}

/* added in v3.11 */
#ifndef __BIN_ATTR
#define __BIN_ATTR(_name, _mode, _read, _write, _size) {    \
   .attr = { .name = __stringify(_name), .mode = _mode },      \
   .read = _read,                \
   .write   = _write,                  \
   .size = _size,                \
}

#define __BIN_ATTR_RO(_name, _size) \
   __BIN_ATTR(_name,       \
   S_IRUGO, _name##_read,  \
   NULL, _size)

#define __BIN_ATTR_RW(_name, _size) \
   __BIN_ATTR(_name,       \
   (S_IWUSR | S_IRUGO), _name##_read,  \
   _name##_write, _size)
#endif

static struct bin_attribute sfp_bin_attrs[] = {
   __BIN_ATTR_RO(bh_sfp_id, 256),
   __BIN_ATTR_RW(bh_sfp_diag, 256),
};

int cpld_sfp_probe(struct device *dev)
{
   unsigned i;
   int err;

   // set programmable timeout in soft i2c
   writeb_cpld_i2c_c5(REG_I2C_C5_TIMEOUT);

   // execute dummy read to initialize i2c start state
   cpld_sfp_read_byte(SFP_BUS_BH, SFP_DEV_ID, 0);

   for (i=0; i<ARRAY_SIZE(sfp_bin_attrs); ++i) {
      /* automatically removed at exit */
      err = device_create_bin_file(dev, &sfp_bin_attrs[i]);
      if (err)
         return err;
   }

   return 0;
}

int cpld_sfp_proc_read(struct seq_file *seq, void *v)
{
   seq_printf(seq, "%-25s = %u\n", "SfpIdCachedUsed", sfp_id_cache.used);

   return 0;
}

