#include <asm/io.h>
#include <asm/page.h>
#include <asm/memory.h>
#include <asm/tlbflush.h>
#include <linux/uaccess.h>
#include <linux/mm.h>
#include <linux/ratelimit.h>
#include <linux/highmem.h>
#include <linux/vmalloc.h>
#include <linux/io.h>

#include "emtimer.h"

static u32 __iomem *free_pool_reg;
static u32 __iomem *submit_reg;
static DEFINE_SPINLOCK(em_ioremap_lock);
static struct vm_struct *em_ioremap_area;

struct dma_mapping {
   void __iomem *virt;
   unsigned long dma;
   unsigned size;
};

static inline void* dma_to_virt(const struct dma_mapping *m, unsigned long dma)
{
   dma -= m->dma;
   if (dma >= m->size) {
      return NULL;
   }
   
   return (void __force *)m->virt + dma;
}

static struct dma_mapping region;
static struct dma_mapping pool;

struct em_descriptor {
   u8  cohMode;
   u8  dirtyFlag;
   u8  dscWsize;
   u8  poolIdx;
   
   u16 freeInfo;
   u8  bufMode;
   u8  eventMode;

   /* start of CPPI descriptor */
   u32 eventSize;

   u8  dstTagLow;
   u8  dstTagHigh;
   u8  srcTagLow;
   u8  srcTagHigh;

   u16 returnInfo;
   u8  psFlags;
   u8  psWsize;

   u32 bufferSize;

   u32 bufferPtr;

   u32 nextDscPtr;

   u32 origBufferSize;

   u32 origBufferPtr;

   u32 timeStamp;

   u16 queueIdx;
   u8  queueType;
   u8  eventType;

   u16 eventGroupIdx;
   u16 reserved;

   u32 freeFunc;

   u32 psWord[];
};
#define QUEUE_MASK 0x3fff
#define BUF_SIZE_MASK ((1<<22)-1)
#define EM_TO_CPPI_OFF offsetof(struct em_descriptor, eventSize)


int __init em_send_init(void)
{
   /* allocate 2 pages for copy_buffer */
   em_ioremap_area = __get_vm_area(2*PAGE_SIZE, VM_IOREMAP, VMALLOC_START, VMALLOC_END);
   if (!em_ioremap_area) {
      return -ENOMEM;
   }

   return 0;
}

static phys_addr_t dma_to_phys(u32 dma)
{
   if (dma < 0x80000000) {
      return dma;
   }else{
      return 0x800000000ULL-0x80000000+dma;
   }
}

static int map_dma_mem(struct dma_mapping *m, unsigned long dma, unsigned size)
{
   phys_addr_t phys = dma_to_phys(dma);
   unsigned long pfn = PFN_DOWN(phys);

   if (pfn_valid(pfn)) {
      m->virt = (void __iomem *)pfn_to_kaddr(pfn)+(dma&~PAGE_MASK);
   } else {
      m->virt = ioremap_cached(phys, size);
   }
   
   if (!m->virt)
      return -EFAULT;
   
   m->dma = dma;
   m->size = size;
   
   return 0;
}

static void unmap_dma_mem(struct dma_mapping *m)
{
   if (!m->virt)
      return;

   if (!virt_addr_valid(m->virt)) {
      iounmap(m->virt);
   }
   
   m->virt = NULL;
}

static void em_send_cleanup(void)
{
   unmap_dma_mem(&pool);

   unmap_dma_mem(&region);

   if (submit_reg) {
      iounmap(submit_reg);
      submit_reg = NULL;
   }

   if (free_pool_reg) {
      iounmap(free_pool_reg);
      free_pool_reg = NULL;
   }
}


int em_send_config(const struct emtimer_config *arg)
{
   em_send_cleanup();

   free_pool_reg = ioremap(0x2a80000 + (arg->free_queue*0x10) + 0xc, 4);
   if (!free_pool_reg) {
      goto out;
   }
   
   submit_reg = ioremap(0x2a80000 + (arg->submit_queue*0x10) + 0xc, 4);
   if (!submit_reg) {
      goto out_free_pool;
   }
   
   if (map_dma_mem(&region, arg->region_dma_addr, arg->region_size)) {
      goto out_submit;
   }
   
   /* also remove the offset from a CPPI descriptor to EM descriptor */
   region.dma += EM_TO_CPPI_OFF;
   region.size -= EM_TO_CPPI_OFF;

   if (map_dma_mem(&pool, arg->pool_dma_addr, arg->pool_size)) {
      goto out_region;
   }
   
   return 0;

out_region:
   unmap_dma_mem(&region);
out_submit:
   iounmap(submit_reg);
   submit_reg = NULL;
out_free_pool:
   iounmap(free_pool_reg);
   free_pool_reg = NULL;
out:
   return -ENOMEM;
}

void __exit em_send_exit(void)
{
   free_vm_area(em_ioremap_area);

   em_send_cleanup();
}

static int copy_buffer(void *dest, phys_addr_t addr, unsigned size)
{
   int err;
   unsigned long vaddr;
   const unsigned long vmap_size = 2*PAGE_SIZE;
   unsigned pageoff;
   
   pageoff = addr&~PAGE_MASK;
   addr &= PAGE_MASK;
   vaddr = (unsigned long)em_ioremap_area->addr;

   spin_lock(&em_ioremap_lock);

   err = ioremap_page_range(vaddr, vaddr+vmap_size, addr, PAGE_KERNEL);
   if (err) {
      spin_unlock(&em_ioremap_lock);
      return err;
   }

   memcpy(dest, (void*)vaddr+pageoff, size);

   flush_cache_vunmap(vaddr, vaddr+vmap_size);
   unmap_kernel_range_noflush(vaddr, vmap_size);
   local_flush_tlb_kernel_range(vaddr, vaddr+vmap_size);
   
   spin_unlock(&em_ioremap_lock);
   
   return 0;
}

int em_send_create(struct emtimer *t, const struct emtimer_create *arg)
{
   struct em_descriptor desc;
   unsigned size;

   /* mask queue id -- this cannot detect bad queues passed in */
   t->queue &= QUEUE_MASK;

   if (copy_from_user(&desc, arg->event-EM_TO_CPPI_OFF, sizeof(desc)))
      return -EFAULT;

   /* validate buffer size */
   t->eventSize = desc.eventSize;
   size = t->eventSize & BUF_SIZE_MASK;
   if (size > sizeof(t->event)) {
      pr_err("emtimer: src event size %u > max (%u)\n",
         size, sizeof(t->event));
      return -EMSGSIZE;
   }
   
   /* copy eventType */
   t->queue |= desc.eventType<<24;
   
   /* copy buffer */
   return copy_buffer(t->event, dma_to_phys(desc.bufferPtr), size);
}

int em_send(const struct emtimer *t)
{
   u32 dma;
   struct em_descriptor *desc;
   void *buf;
   unsigned bufsz, size;

   pr_debug("em_send(%08x, %p)\n", t->queue, t->event);
   
   /* pop desc */
retry:
   dma = __raw_readl(free_pool_reg);
   if (!dma) {
      pr_err_ratelimited("emtimer: no free descriptors\n");
      return -ENOBUFS;
   }
   
   desc = dma_to_virt(&region, dma&~0xf);
   if (!desc) {
      /* drop descriptor */
      //return -EFAULT;
      goto retry;
   }

   /* destination space check */
   bufsz = desc->origBufferSize & BUF_SIZE_MASK;
   size = t->eventSize & BUF_SIZE_MASK;
   if (bufsz < size) {
      /* drop descriptor */
      //return -ENOSPC;
      goto retry;
   }
   
   buf = dma_to_virt(&pool, desc->origBufferPtr);
   if (!buf) {
      /* drop descriptor */
      //return -EFAULT;
      goto retry;
   }
   memcpy(buf, t->event, size);

   desc->eventSize = t->eventSize;
   desc->psWsize = 0x80; /* EPIB present, no psWords */
   desc->bufferSize = size;
   desc->bufferPtr = desc->origBufferPtr;
   desc->queueIdx = t->queue;
   desc->eventType = t->queue>>24;

   /* push */
   /* execute a dsb(st) to prevent reordering between normal/device memory */
   writel(dma, submit_reg);
   
   return 0;
}
