/*
 * Copyright (c) 2014 Nokia Solutions and Networks Inc
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */

/**
 * @file fpga-bus.c
 * @author Nokia Solutions and Networks
 * @remarks Copyright (C) 2014 Nokia Solutions and Networks
 *
 * @par
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * @par PURPOSE
 * FPGA bus and codeload driver
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    SRID       Description
 * ---------- --------- ---------- ----------------------------------------------
 * 11/23/2010 qdpx86    1372454    Initial creation from mod_cpld
 * 07/29/2011 qdpx86    1426680    Port to ZeNB AP
 * 02/07/2012 qdpx86    1458933    Checking memory allocation
 * 02/21/2012 qdpx86    1458933    Updates for Altera codeload
 * 03/06/2012 qdpx86    1458933    Adding bitswap for Altera
 * 05/02/2012 tberger1  1458932    Skip FPGA codeload for SoC only reset
 * 02/03/2015 tberger1             Remove bitswap, Altera FPGA images
 *                                 will be bitswapped at FZM build time
 * </pre>
 */
#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/device.h>
#include <linux/platform_device.h>
#include <linux/firmware.h>
#include <linux/slab.h>
#include <linux/jiffies.h>
#include <linux/delay.h>
#include <linux/cdev.h>
#include <linux/fs.h>
#include <linux/of.h>
#include <linux/spi/spi.h>
#include <asm/uaccess.h>
#include <asm/io.h>
#include "fzm_fpga.h"
#include "fzm_fpga_eis.h"
#include "fzm_cpld.h"
#include "fpga_internal.h"

static int fpga_loaded;

static struct device *fpgadev;

static int fpga_bus_match(struct device *dev, struct device_driver *drv)
{
   struct fpga_device *fdev = to_fpga_device(dev);

   /* don't match anything until the fpga has been loaded */
   return fpga_loaded && (strcmp(fdev->name, drv->name) == 0);
}

static int fpga_bus_uevent(struct device *dev, struct kobj_uevent_env *env)
{
   if (!dev->of_node)
      return -ENODEV;

   return add_uevent_var(env, "MODALIAS=%s", dev->of_node->name);
}

static int fpga_bus_device_probe(struct device *_dev)
{
   struct fpga_driver *drv = to_fpga_driver(_dev->driver);
   struct fpga_device *dev = to_fpga_device(_dev);

   return drv->probe(dev);
}

static int fpga_bus_device_remove(struct device *_dev)
{
   struct fpga_driver *drv = to_fpga_driver(_dev->driver);
   struct fpga_device *dev = to_fpga_device(_dev);

   return drv->remove(dev);
}

static struct bus_type fpga_bus_type = {
   .name = "fpga_bus",
   .match = fpga_bus_match,
   .uevent = fpga_bus_uevent,
   .probe = fpga_bus_device_probe,
   .remove = fpga_bus_device_remove,
};

static void fpga_device_release(struct device *dev)
{
   struct fpga_device *fdev = to_fpga_device(dev);

   kfree(fdev);
}

/**
 * Adds a device to the FPGA bus.  The name of the device will be matched
 * with the name of a fpga driver after codeload is completed, in order to
 * determine what driver should handle the device.
 *
 * @param name Name of device, which should match driver name
 * @param id unique id of device, or -1 if there is only 1 device of this type
 * @param data data field to set in device, to be used by driver.  If this
 *   is a pointer, it should be to static or dynamically allocated memory.
 *
 * @return fpga_device structure, or ERR_PTR on failure (use IS_ERR and PTR_ERR)
 */
struct fpga_device *fpga_device_register(const char *name, int id, void *data, struct device_node *node)
{
   struct fpga_device *fdev;
   int err;

   fdev = kzalloc(sizeof(*fdev) + strlen(name) + 1, GFP_KERNEL);
   if (!fdev) {
      return ERR_PTR(-ENOMEM);
   }

   fdev->dev.parent = fpgadev;
   fdev->dev.bus = &fpga_bus_type;
   fdev->dev.platform_data = data;
   fdev->dev.release = fpga_device_release;

   if (node) {
      of_node_get(node);
      fdev->dev.of_node = node;
   }

   fdev->id = id;
   strcpy(fdev->name, name);

   //use ID as well to avoid duplicates
   if (id != -1)
      dev_set_name(&fdev->dev, "%s.%u", name, id);
   else
      dev_set_name(&fdev->dev, "%s", name);

   err = device_register(&fdev->dev);
   if (err)
      goto out;

   return fdev;
out:
   kfree(fdev);
   return ERR_PTR(-ENOMEM);
}
EXPORT_SYMBOL(fpga_device_register);

static void fpga_cdev_release(struct device *dev)
{
   cdev_del(dev->platform_data);

   kfree(dev);
}

static struct class fpga_cdev_class = {
      .name = "fpga_cdev",
      .owner = THIS_MODULE,
      .dev_release = fpga_cdev_release,
   };

/**
 * Creates a character device node associated with the fpga device.
 * A uevent will be sent to udev/mdev to notify of the cdev creation.
 *
 * @param devt device node to use, previously allocated with alloc_chrdev_region
 * @param fops file operations structure for character device
 * @param parent fpga_device which owns the character device
 * @param name format string for chardev node name
 * @param ... parameters for name format string
 *
 * @return 0 on success, -errno on failure.
 */
int fpga_cdev_create(dev_t devt, const struct file_operations *fops, struct fpga_device *parent, const char *name, ...)
{
   struct device *dev;
   struct cdev *cdev;
   va_list vargs;
   int err;
   char namebuf[64];

   dev = kzalloc(sizeof(*dev) + sizeof(*cdev), GFP_KERNEL);
   if (!dev) {
      return -ENOMEM;
   }

   device_initialize(dev);

   cdev = (void *)(dev + 1);
   cdev_init(cdev, fops);

   dev->devt = devt;
   dev->parent = &parent->dev;
   dev->platform_data = cdev;
   dev->class = &fpga_cdev_class;

   va_start(vargs, name);
   vsnprintf(namebuf, sizeof(namebuf), name, vargs);
   va_end(vargs);

   err = dev_set_name(dev, "%s", namebuf);
   if (err)
      goto out_cdev;

   err = cdev_add(cdev, devt, 1);
   if (err)
      goto out_cdev;

   err = device_add(dev);
   if (err)
      goto out_cdev;

   return 0;

out_cdev:
   cdev_del(cdev);
   put_device(dev);

   return err;
}
EXPORT_SYMBOL(fpga_cdev_create);

static int fpga_cdev_cleanup_cb(struct device *dev, void *data)
{
   /* release function will cleanup cdev */
   if (dev->class == &fpga_cdev_class)
      device_unregister(dev);

   return 0;
}

/**
 * Removes and cleans up all character devices associated with a fpga device.
 *
 * @param parent fpga_device which owns the character devices
 */
void fpga_cdev_cleanup(struct fpga_device *parent)
{
   device_for_each_child(&parent->dev, NULL, fpga_cdev_cleanup_cb);
}
EXPORT_SYMBOL(fpga_cdev_cleanup);


/**
 * Adds a driver to the FPGA bus.  The name of the driver will be matched
 * with the name of fpga driver(s) after codeload is completed.
 *
 * @param drv Driver structure with at least .driver.{name,owner}, .probe, and .remove initalized
 *
 * @return 0 on success, -errno on failure.
 */
int fpga_driver_register(struct fpga_driver *drv)
{
   drv->driver.bus = &fpga_bus_type;
   return driver_register(&drv->driver);
}
EXPORT_SYMBOL(fpga_driver_register);

static void fpga_loaded_set(int loaded)
{
   int err;

   if (loaded != fpga_loaded) {
      fpga_loaded = loaded;

      //dev_info(fpgadev, "FPGA version 0x%08x is loaded\n", fpga_read(0));
      dev_info(fpgadev, "FPGA version is loaded\n");

      /* rescan since the fpga is loaded now */
      err = bus_rescan_devices(&fpga_bus_type);
      if (err)
         dev_err(fpgadev, "rescan devices failed: %d\n", err);
   }
}

static int fpga_release_reset(void)
{
   cpld_lock();
   writeb_cpld_rrr2(readb_cpld_rrr2() | CPLD_RRR2_BIT_FPGA);
   cpld_unlock();

   return 0;
}

static int fpga_reset(void)
{
   cpld_lock();
   writeb_cpld_rrr2(readb_cpld_rrr2() & ~CPLD_RRR2_BIT_FPGA);
   cpld_unlock();

   return 0;
}

//PROG == CONFIG
//STATUS == INIT

/* 1) Pull PROG line low for 500ns.
 * 2) Pull PROG line high.
 * 3) Wait for INIT line to go high.
 * 4) Write to MASK once, and then DATA register for each byte.
 * 5) Adopt best practice to ensure configuration of all possible bit
 * patterns.  This is achieved by writing all 1s of data while waiting
 * for DONE bit to go high.
 * 6) When DONE bit goes high, write 8 more CCLK cycles of all 1s.  (A
 * CCLK cycle is a write to the DATA register.
 * 7) Clear the MASK register.
 */
static int codeload_begin(void)
{
   int err;
   unsigned long timeout;

   fpga_loaded_set(0);

   /* set chip select to begin */
   //writeb_cpld_mask(CPLD_MASK_BIT_MASK);

   /* deassert nCONFIG (0) */
   writeb_cpld_prog(readb_cpld_prog() & ~CPLD_PROG_BIT_PROG);

   /* ensure nCONFIG is low for at least 500ns */
   ndelay(500);

   /* wait for nSTATUS low (0) */
   timeout = jiffies + HZ;
   while ((readb_cpld_init() & CPLD_INIT_BIT_INIT)) {
      if (time_after(jiffies, timeout)) {
         err = -ETIMEDOUT;
         dev_err(fpgadev, "Did not get nSTATUS low\n");
         goto out;
      }
      cpu_relax();
   }

   /* assert nCONFIG (1) */
   writeb_cpld_prog(readb_cpld_prog() | CPLD_PROG_BIT_PROG);

   /* wait for nSTATUS to go high again */
   timeout = jiffies + HZ;
   while (!(readb_cpld_init() & CPLD_INIT_BIT_INIT)) {
      if (time_after(jiffies, timeout)) {
         err = -ETIMEDOUT;
         dev_err(fpgadev, "Did not get nSTATUS high\n");
         goto out;
      }
      cpu_relax();
   }

   dev_dbg(fpgadev, "got init, init=%u done=%u\n", readb_cpld_init(), readb_cpld_done());

   return 0;

out:
   /* deassert chip select */
   //writeb_cpld_mask(0);

   return err;
}

static int codeload_block(const u8 *ptr, size_t count)
{
   /* Note: bitswapping for Altera not needed here
    * because Altera FPGA images will be pre-processed
    * during the FZM image build to reverse the bits per byte */
   fzm_cpldspi_write(FZM_CPLDSPI_FPGA_CONFIG, 0, (void *)ptr, count);

   if (!(readb_cpld_init() & CPLD_INIT_BIT_INIT)) {
      dev_err(fpgadev, "nSTATUS went low during program!\n");
      return -EIO;
   }

   return 0;
}

static int codeload_end(void)
{
   int err = 0;
   unsigned long timeout;
   u8 dummybuf[8] = {0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff};

   if (!(readb_cpld_init() & CPLD_INIT_BIT_INIT)) {
      return -EIO;
   }

   dev_dbg(fpgadev, "polling on done. init=%u\n", readb_cpld_init());

   /* 5 */
   /* Continue to write all 1s into DATA until DONE goes high */
   timeout = jiffies + (5 * HZ);
   while (!(readb_cpld_done() & CPLD_DONE_BIT_DONE)) {
      if (time_after(jiffies, timeout)) {
         err = -ETIMEDOUT;
         dev_err(fpgadev, "Did not get DONE signal\n");
         goto out;
      }
      fzm_cpldspi_write(FZM_CPLDSPI_FPGA_CONFIG, 0, dummybuf, 1);
      cpu_relax();
   }

   //need at least 2 additional clocks
   fzm_cpldspi_write(FZM_CPLDSPI_FPGA_CONFIG, 0, dummybuf, 8);

   //FIXME?
   //charlie waits 150us before bringing out of reset
   udelay(150);

out:
   /* deassert chip select */
   //writeb_cpld_mask(0);

   return err;
}

static int fpga_codeload(void)
{
   const struct firmware *fw = NULL;
   int err;

   //printk("<3>Codeload Probe\n");

   /* skip codeload if previous reset was a soft reset and we can read fpga */
   /*
   if ( (1 == dan_softreset_prev_get()) && (0xffffffff != fpga_read(0)) ) {
      dev_info(fpgadev->dev, "skipping FPGA codeload\n");
   } else {
   */
   err = request_firmware(&fw, "fpga.bin", fpgadev);
   if (err)
      goto out;

   //printk("<3>Firmware found\n");

   err = codeload_begin();
   if (err)
      goto out_fw;

   //printk("<3>Codeload begun\n");

   err = codeload_block(fw->data, fw->size);
   if (err)
      goto out_fw;

   //printk("<3>Code block wrote\n");

   err = codeload_end();
   if (err)
      goto out_fw;

   //printk("<3>Code load ended\n");
   /* the FPGA must be out of reset before it can be accessed */
   err = fpga_release_reset();
   if (err)
      goto out_fw;
   msleep(500);
   /*} */

   //printk("<3>FPGA released from reset\n");

   /*
   err = fpga_ic_probe(&pdev->dev);
   if (err)
      goto out_fw;
   */

   /* let all other drivers probe */
   fpga_loaded_set(1);

   return 0;

out_fw:
   if (fw)
      release_firmware(fw);
out:
   return err;
}

void fpga_bus_register_devices(struct device *parent)
{
   struct device_node *node;

   for_each_available_child_of_node(parent->of_node, node) {
      struct fpga_device *cdev;
      int child_id = -1;
      u32 tmp_id;

      if (!of_property_read_u32(node, "reg", &tmp_id))
         child_id = tmp_id;

      request_module_nowait(node->name);

      cdev = fpga_device_register(node->name, child_id, NULL, node);
      if (IS_ERR(cdev)) {
         dev_err(parent, "cannot register child %s\n", node->name);
         continue;
      }
   }
}

static int fpga_bus_clean(struct device *dev, void *arg)
{
   device_unregister(dev);
   return 0;
}

void fpga_bus_unregister_devices(struct device *parent)
{
   bus_for_each_dev(&fpga_bus_type, NULL, NULL, fpga_bus_clean);
}

int fpga_bus_probe(struct device *dev)
{
   int err;

   err = bus_register(&fpga_bus_type);
   if (err)
      return err;

   err = class_register(&fpga_cdev_class);
   if (err)
      goto out_bus;

   fpgadev = dev;

   err = fpga_codeload();
   if (err)
      goto out_class;

   return 0;

out_class:
   class_unregister(&fpga_cdev_class);
out_bus:
   bus_unregister(&fpga_bus_type);

   return err;
}

int fpga_bus_remove(struct device *dev)
{
   fpga_reset();
   class_unregister(&fpga_cdev_class);
   bus_unregister(&fpga_bus_type);

   return 0;
}
