/*
 * Copyright (c) 2005-2009 Motorola Inc
 * Copyright (c) 2014 Nokia Solutions and Networks
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. The name of the author may not be used to endorse or promote products
 *    derived from this software without specific prior written permission.
 *
 * Alternatively, this software may be distributed under the terms of the
 * GNU General Public License ("GPL") version 2 as published by the Free
 * Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 */
/**
 * @file mod_utils.h
 * @author Motorola / Nokia Solutions and Networks
 * @remarks Copyright (C) 2005-2009 Motorola Inc, 2014 Nokia Solutions and Networks
 *
 * @par PURPOSE Common macros for kernel modules.
 *
 * @par REVISIONS
 * <pre>
 * Date       CoreID    DDTSID     Description
 * ---------- --------- ---------- --------------------------------------------
 * 12/28/2005 Motorola             Initial Version
 * 07/29/2011 NSN                  Reused/updated for ZeNB AP.
 * </pre>
 */

#ifndef MOD_UTILS_H
#define MOD_UTILS_H

/* The following macros are to be used when printing non-error related
 * information. This includes function tracing as well as general debugging
 * messages.  This allows for them to all be turned on and off with a single
 * build-time define value.
 *
 * MOD_DEBUG_LEVEL=3
 * Everything is enabled. This includes function tracing (MOD_TRACE),
 * debugging (MOD_DEBUG), and information (MOD_INFO).
 *
 * MOD_DEBUG_LEVEL=2
 * Debugging (MOD_DEBUG) and information (MOD_INFO) are enabled.
 *
 * MOD_DEBUG_LEVEL=1
 * Information (MOD_INFO) is enabled.
 *
 * MOD_DEBUG_LEVEL=0 or not defined at all
 * Nothing is enabled.
 *
 * MOD_INFO should be used wherever you think general information would be
 * useful, such as a module initialization completing.
 *
 * MOD_DEBUG should be used for general debugging information such as a
 * specific if-branch being taken.
 *
 * MOD_TRACE should be used at the top of each function so that when function
 * tracing is enabled, it can be viewed in the log.
 */

#define MOD_CRIT(f, x...)  pr_alert("%s(): "f, __FUNCTION__, ##x)

#ifdef MOD_DEBUG_LEVEL
/* MOD_TRACE is enabled at level 3 */
#  if (MOD_DEBUG_LEVEL>2)
#define MOD_TRACE(x)              printk(KERN_DEBUG "TRACE:" #x "\n")
#define MOD_TRACE1(x, a1)         printk(KERN_DEBUG "TRACE:" #x " %s=0x%08X\n", #a1, (unsigned int)(a1))
#define MOD_TRACE2(x, a1, a2)     printk(KERN_DEBUG "TRACE:" #x " %s=0x%08X %s=0x%08X\n", #a1, (unsigned int)(a1), #a2, (unsigned int)(a2))
#define MOD_TRACE3(x, a1, a2, a3) printk(KERN_DEBUG "TRACE:" #x " %s=0x%08X %s=0x%08X %s=0x%08X\n", #a1, (unsigned int)(a1), #a2, (unsigned int)(a2), \
      #a3, (unsigned int)(a3))
#define MOD_DEBUG(f, x...) printk(KERN_DEBUG "%s(): "f,__FUNCTION__,##x)
#define MOD_ERROR(f, x...) printk(KERN_ERR  "%s(): "f,__FUNCTION__,##x)
#define MOD_INFO(f, x...)  printk(KERN_INFO "%s(): "f,__FUNCTION__,##x)
#  elif (MOD_DEBUG_LEVEL>1)
#define MOD_TRACE(x)
#define MOD_TRACE1(x, a1)
#define MOD_TRACE2(x, a1, a2)
#define MOD_TRACE3(x, a1, a2, a3)
#define MOD_DEBUG(f, x...) printk(KERN_DEBUG "%s(): "f,__FUNCTION__,##x)
#define MOD_ERROR(f, x...) printk(KERN_ERR  "%s(): "f,__FUNCTION__,##x)
#define MOD_INFO(f, x...)  printk(KERN_INFO "%s(): "f,__FUNCTION__,##x)
#  elif (MOD_DEBUG_LEVEL>0)
#define MOD_TRACE(x)
#define MOD_TRACE1(x, a1)
#define MOD_TRACE2(x, a1, a2)
#define MOD_TRACE3(x, a1, a2, a3)
#define MOD_DEBUG(x...)
#define MOD_INFO(f, x...)  printk(KERN_INFO "%s(): "f,__FUNCTION__,##x)
#define MOD_ERROR(f, x...) printk(KERN_ERR  "%s(): "f,__FUNCTION__,##x)
#  else
#define MOD_TRACE(x)
#define MOD_TRACE1(x, a1)
#define MOD_TRACE2(x, a1, a2)
#define MOD_TRACE3(x, a1, a2, a3)
#define MOD_ERROR(f, x...)
#define MOD_DEBUG(x...)
#define MOD_INFO(x...)
#  endif
#else /* MOD_DEBUG_LEVEL wasn't defined at all */
#define MOD_TRACE(x)
#define MOD_TRACE1(x, a1)
#define MOD_TRACE2(x, a1, a2)
#define MOD_TRACE3(x, a1, a2, a3)
#define MOD_ERROR(f, x...)
#define MOD_DEBUG(x...)
#define MOD_INFO(x...)
#endif

#endif /* MOD_UTILS_H */
