/*****************************************************************************
 * Project Name: FSMr3
 * Module Name:  message dispatcher
 * File Name:    se_cfg_mod.c
 *
 * Author:        Junshu Zhang
 * Creation Date: 03 Aug 2010
 * Last modified: 20 Aug 2010
 * ---------------------------------------------------------------------------
 * Description: Kernel driver for exchanging configuration/perf.data/debug
 *              information between user space and SEs
 *****************************************************************************
 * Copyright (c) Nokia 2010, all rights reserved
 * Licensed material - property of Nokia 
 *****************************************************************************
 * Revision history (latest on top)
 *
 * Date        | Who | Description
 * ------------+-----+--------------------------------------------------------
 * 04-Nov-2010 | JZh | updated process request
 *                     added counters and debug flags (/sys/kernel/msgdp)
 * 02-Sep-2010 | DSi | Updated TARGET_SE_IP_POW_TAG value to 0xFEED
 *                     Added check for IP configuration, not to wait for ret msg
 *                     from SE.
 * 20-Aug-2010 | JZh | generic support for ATM and IPSec SEs
 * 03-Aug-2010 | JZh | init version to support ATM config
 * 25-Jun-2010 | GFr | init version echo reply
 * 08-Sep-2015 | AJo | Moved Kernel module adaptation to FZM specific IPCi WoW.
 *****************************************************************************/

/******************************* INCLUDES ********************************/
#include <linux/kobject.h>
#include <linux/sysfs.h>
#include <linux/module.h>
#include <linux/init.h>
#include <linux/sched.h>
#include <linux/delay.h>
#include <linux/jiffies.h>
#include <net/netlink.h>
#include <net/genetlink.h>
#include <linux/string.h>
#include "ip_secfg_IWF_VTC.h"
#include "seCfgDrv_If.h"
#include "se_cfg_mod_help.h"

/************************ FUNCTION DECLARATIONS *************************/

int seCfgDrv_init(void);
void seCfgDrv_exit(void);
/* msg processing */
//static int process_echo_request(struct sk_buff* skb, struct genl_info* info);
static void process_request(struct sk_buff* skb);
static int process_reply(char* data);

#define IPCi_ERR  -1
#define IPCi_OK  0

/****************************** VARIABLES *******************************/
#define FPCI_MAX_DATA_SIZ           2048 
struct sock *nl_sk;

typedef struct _ipci_dbgcnt_t 
{
    unsigned ct_nl_rx;
    unsigned ct_nl_rx_len_err;
    unsigned ct_nl_tx;
    unsigned ct_nl_tx_err;
    unsigned ct_kr_tx_err;
}ipci_dbgcnt_t;

ipci_dbgcnt_t ipciDbgCnt; //TODO : create /proc fs entry to dump these. 

/************************ FUNCTION IMPLEMENTATION ***********************/
int seCfgDrv_init(void)
{
    int retVal = 0;
    struct netlink_kernel_cfg nl_cfg =
    {
         .input = process_request,
    };
    memset(&ipciDbgCnt,0, sizeof(ipci_dbgcnt_t));
   
    /* Register new NSN specific netlink family */
    nl_sk=netlink_kernel_create(&init_net, NETLINK_USERSOCK, &nl_cfg);
    if(!nl_sk)
    {
        printk(KERN_INFO "seCfgDrv_init: netlink_kernel_create failed\n");
        retVal = -1;
    }
    return retVal;
}

void seCfgDrv_exit(void)
{
   netlink_kernel_release(nl_sk); 
}

static void  process_request(struct sk_buff* skb)
{
    struct nlmsghdr *nlh;
    int error_msg = EM_SUCCESS;
    ipciDbgCnt.ct_nl_rx++;
    if(NULL != skb->data)
    {
        nlh=(struct nlmsghdr*)skb->data;
        if(FPCI_MAX_DATA_SIZ >= nlh->nlmsg_len)
        {
            error_msg = ip_se_api(skb->data);
            if(EM_SUCCESS != error_msg)
            {
                ipciDbgCnt.ct_kr_tx_err++;
                printk(KERN_INFO " Error returned \n");
            }
        }
        else
        {
            error_msg = IPCi_ERR;
            ipciDbgCnt.ct_nl_rx_len_err++;
            printk(KERN_INFO "NL length exceeded limit \n");
        }
        if(0 != error_msg)
        {
            sKernelIpCfgMsgHeader *pIpCfgMsg = (sKernelIpCfgMsgHeader *)(skb->data);
            pIpCfgMsg->retCode = IPCi_ERR;
        }    
    
        /* reply to user space */
        error_msg  = process_reply(skb->data);
        if(error_msg != IPCi_OK )
        {
            printk(KERN_INFO "Error process_reply returns %d\n", error_msg);
        }
    }
    else
    {
        
        printk(KERN_WARNING "process_request: Invalid data received in Kernel module\n");
    }
}

static int process_reply(char* data)
{

    int retVal = IPCi_OK;
    struct sk_buff *skb_out;
    struct nlmsghdr *nlh, *nlh_new;
    int data_siz = 0;

    nlh=(struct nlmsghdr*)data;
    data_siz = nlh->nlmsg_len - sizeof(struct nlmsghdr);

    if(0 < data_siz)
    {
        skb_out = nlmsg_new(data_siz, GFP_ATOMIC);
        if(!skb_out)
        {
            retVal = IPCi_ERR;
        }
        else
        {
            nlh_new=nlmsg_put(skb_out,0,nlh->nlmsg_seq,NLMSG_DONE,data_siz,0);
            memcpy(nlmsg_data(nlh_new), NLMSG_DATA(nlh), data_siz);
            NETLINK_CB(skb_out).dst_group = 0; 
            retVal = nlmsg_unicast(nl_sk, skb_out, nlh->nlmsg_pid);

            if(IPCi_OK != retVal)
            {
                ipciDbgCnt.ct_nl_tx_err++;
                printk(KERN_INFO "process_reply: nlmsg_unicast, pid=%d\n", nlh->nlmsg_pid);
            }
            else
            {
                ipciDbgCnt.ct_nl_tx++;
            }
        }
    }
    else
    {
        retVal = IPCi_ERR;
        printk(KERN_INFO "nsn_grp_nl_send_msg: Invalid data size = %d\n", data_siz);
    }
    return retVal;
}
#if 0 
static int process_echo_request(struct sk_buff* skb, struct genl_info* info)
{
    int rc = 0;

    struct nlattr* myAttr = NULL;
    struct sk_buff* replySkb = NULL;
    char* echoData = NULL;
    void* msgHead = NULL;
    char replyString[SECFG_MAX_STRING_LEN];
    int echoAllocated = 0;

    if (NULL == info)
    {
        printk(KERN_WARNING "Empty info structure!\n");
        rc = -EINVAL;
        goto func_exit;
    }

    /* echo request should have a A_SECFG_ECHO_STRING attribute */
    myAttr = info->attrs[A_SECFG_ECHO_STRING];
    if (myAttr)
    {
        echoData = (char*)nla_data(myAttr);
        if ((NULL == echoData) || (!strcmp(echoData, "")))
        {
            printk(KERN_WARNING "error receiving data");
            echoData = kstrdup("<empty>",  GFP_KERNEL);
            echoAllocated = 1;
        }
    }
    else
    {
        printk(KERN_WARNING "No A_SECFG_ECHO_STRING attribute found!\n");
        rc = -EINVAL;
        goto func_exit;
    }


    /* compose reply message */
    replySkb = genlmsg_new(NLMSG_GOODSIZE, GFP_KERNEL);
    if (NULL == replySkb)
    {
        printk(KERN_ERR "Could not allocate socket buffer for reply!\n");
        rc = -ENOMEM;
        goto func_exit;
    }

    msgHead = genlmsg_put_reply(replySkb, info, &seCfg_genl_family, 0,
            C_SECFG_ECHO_REPLY);
    if (NULL == msgHead)
    {
        printk(KERN_ERR "Error composing reply message header!\n");
        rc = -ENOMEM;
        goto func_exit;
    }

    strncpy(replyString, "Reply: ", SECFG_MAX_STRING_LEN);
    strncat(replyString, echoData, SECFG_MAX_STRING_LEN);
    rc = nla_put_string(replySkb, A_SECFG_ECHO_STRING, replyString);
    if (rc != 0)
    {
        printk(KERN_WARNING "Error adding reply string, rc is %d\n", rc);
        goto func_exit;
    }

    genlmsg_end(replySkb, msgHead);

    rc = genlmsg_reply(replySkb, info);
    if (rc != 0)
    {
        printk(KERN_WARNING "Error sending reply, rc is %d\n", rc);
        goto func_exit;
    }

    func_exit:
    if (echoAllocated)
    {
        kfree(echoData);
        echoAllocated = 0;
    }

    return rc;
}
#endif
/* formal stuff for kernel modules */

MODULE_AUTHOR("Nokia Siemens Networks");
MODULE_DESCRIPTION("Kernel module to dispatch SE configuration messages");
MODULE_LICENSE("GPL");
